@@ -23,6 +23,7 @@ from time import time
 from urllib.parse import parse_qs, unquote, urlparse
 
 from cli_helpers.tabular_output import TabularOutputFormatter, preprocessors
+from cli_helpers.tabular_output.output_formatter import MISSING_VALUE as DEFAULT_MISSING_VALUE
 from cli_helpers.utils import strip_ansi
 import click
 from configobj import ConfigObj
@@ -37,7 +38,8 @@ from prompt_toolkit.key_binding.key_processor import KeyPressEvent
 from prompt_toolkit.layout.processors import ConditionalProcessor, HighlightMatchingBracketProcessor
 from prompt_toolkit.lexers import PygmentsLexer
 from prompt_toolkit.shortcuts import CompleteStyle, PromptSession
-from pymysql import OperationalError, err
+import pymysql
+from pymysql.constants.ER import HANDSHAKE_ERROR
 from pymysql.cursors import Cursor
 import sqlglot
 import sqlparse
@@ -153,6 +155,15 @@ class MyCli:
         self.destructive_warning = c_dest_warning if warn is None else warn
         self.login_path_as_host = c["main"].as_bool("login_path_as_host")
         self.post_redirect_command = c['main'].get('post_redirect_command')
+        self.null_string = c['main'].get('null_string')
+
+        # set ssl_mode if a valid option is provided in a config file, otherwise None
+        ssl_mode = c["main"].get("ssl_mode", None)
+        if ssl_mode not in ("auto", "on", "off", None):
+            self.echo(f"Invalid config option provided for ssl_mode ({ssl_mode}); ignoring.", err=True, fg="red")
+            self.ssl_mode = None
+        else:
+            self.ssl_mode = ssl_mode
 
         # read from cli argument or user config file
         self.auto_vertical_output = auto_vertical_output or c["main"].as_bool("auto_vertical_output")
@@ -213,7 +224,7 @@ class MyCli:
     def register_special_commands(self) -> None:
         special.register_special_command(self.change_db, "use", "\\u", "Change to a new database.", aliases=["\\u"])
         special.register_special_command(
-            self.change_db,
+            self.manual_reconnect,
             "connect",
             "\\r",
             "Reconnect to the database. Optional database argument.",
@@ -260,6 +271,18 @@ class MyCli:
             self.change_prompt_format, "prompt", "\\R", "Change prompt format.", aliases=["\\R"], case_sensitive=True
         )
 
+    def manual_reconnect(self, arg: str = "", **_) -> Generator[tuple, None, None]:
+        """
+        Interactive method to use for the \r command, so that the utility method
+        may be cleanly used elsewhere.
+        """
+        if not self.reconnect(database=arg):
+            yield (None, None, None, "Not connected")
+        elif not arg or arg == '``':
+            yield (None, None, None, None)
+        else:
+            yield self.change_db(arg).send(None)
+
     def enable_show_warnings(self, **_) -> Generator[tuple, None, None]:
         self.show_warnings = True
         msg = "Show warnings enabled."
@@ -300,13 +323,18 @@ class MyCli:
             return
 
         assert isinstance(self.sqlexecute, SQLExecute)
-        self.sqlexecute.change_db(arg)
+
+        if self.sqlexecute.dbname == arg:
+            msg = f'You are already connected to database "{self.sqlexecute.dbname}" as user "{self.sqlexecute.user}"'
+        else:
+            self.sqlexecute.change_db(arg)
+            msg = f'You are now connected to database "{self.sqlexecute.dbname}" as user "{self.sqlexecute.user}"'
 
         yield (
             None,
             None,
             None,
-            f'You are now connected to database "{self.sqlexecute.dbname}" as user "{self.sqlexecute.user}"',
+            msg,
         )
 
     def execute_from_file(self, arg: str, **_) -> Iterable[tuple]:
@@ -524,7 +552,7 @@ class MyCli:
                     ssh_key_filename,
                     init_command,
                 )
-            except OperationalError as e:
+            except pymysql.OperationalError as e:
                 if e.args[0] == ERROR_CODE_ACCESS_DENIED:
                     if password_from_file is not None:
                         new_passwd = password_from_file
@@ -549,6 +577,24 @@ class MyCli:
                         ssh_key_filename,
                         init_command,
                     )
+                elif e.args[0] == HANDSHAKE_ERROR and ssl is not None and ssl.get("mode", None) == "auto":
+                    self.sqlexecute = SQLExecute(
+                        database,
+                        user,
+                        passwd,
+                        host,
+                        int_port,
+                        socket,
+                        charset,
+                        use_local_infile,
+                        None,
+                        ssh_user,
+                        ssh_host,
+                        int(ssh_port) if ssh_port else None,
+                        ssh_password,
+                        ssh_key_filename,
+                        init_command,
+                    )
                 else:
                     raise e
 
@@ -558,7 +604,7 @@ class MyCli:
                 self.echo(f"Connecting to socket {socket}, owned by user {socket_owner}", err=True)
                 try:
                     _connect()
-                except OperationalError as e:
+                except pymysql.OperationalError as e:
                     # These are "Can't open socket" and 2x "Can't connect"
                     if [code for code in (2001, 2002, 2003) if code == e.args[0]]:
                         self.logger.debug("Database connection failed: %r.", e)
@@ -774,6 +820,7 @@ class MyCli:
                     headers,
                     special.is_expanded_output(),
                     special.is_redirected(),
+                    self.null_string,
                     max_width,
                 )
 
@@ -806,6 +853,7 @@ class MyCli:
                             headers,
                             special.is_expanded_output(),
                             special.is_redirected(),
+                            self.null_string,
                             max_width,
                         )
                         self.echo("")
@@ -911,19 +959,12 @@ class MyCli:
                 output_res(res, start)
                 special.unset_once_if_written(self.post_redirect_command)
                 special.flush_pipe_once_if_written(self.post_redirect_command)
-            except err.InterfaceError:
-                logger.debug("Attempting to reconnect.")
-                self.echo("Reconnecting...", fg="yellow")
-                try:
-                    sqlexecute.connect()
-                    logger.debug("Reconnected successfully.")
-                    one_iteration(text)
-                    return  # OK to just return, cuz the recursion call runs to the end.
-                except OperationalError as e2:
-                    logger.debug("Reconnect failed. e: %r", e2)
-                    self.echo(str(e2), err=True, fg="red")
-                    # If reconnection failed, don't proceed further.
+            except pymysql.err.InterfaceError:
+                # attempt to reconnect
+                if not self.reconnect():
                     return
+                one_iteration(text)
+                return  # OK to just return, cuz the recursion call runs to the end.
             except EOFError as e:
                 raise e
             except KeyboardInterrupt:
@@ -954,21 +995,14 @@ class MyCli:
                     self.echo("Did not get a connection id, skip cancelling query", err=True, fg="red")
             except NotImplementedError:
                 self.echo("Not Yet Implemented.", fg="yellow")
-            except OperationalError as e1:
+            except pymysql.OperationalError as e1:
                 logger.debug("Exception: %r", e1)
                 if e1.args[0] in (2003, 2006, 2013):
-                    logger.debug("Attempting to reconnect.")
-                    self.echo("Reconnecting...", fg="yellow")
-                    try:
-                        sqlexecute.connect()
-                        logger.debug("Reconnected successfully.")
-                        one_iteration(text)
-                        return  # OK to just return, cuz the recursion call runs to the end.
-                    except OperationalError as e2:
-                        logger.debug("Reconnect failed. e: %r", e2)
-                        self.echo(str(e2), err=True, fg="red")
-                        # If reconnection failed, don't proceed further.
+                    # attempt to reconnect
+                    if not self.reconnect():
                         return
+                    one_iteration(text)
+                    return  # OK to just return, cuz the recursion call runs to the end.
                 else:
                     logger.error("sql: %r, error: %r", text, e1)
                     logger.error("traceback: %r", traceback.format_exc())
@@ -1040,6 +1074,58 @@ class MyCli:
             if not self.less_chatty:
                 self.echo("Goodbye!")
 
+    def reconnect(self, database: str = "") -> bool:
+        """
+        Attempt to reconnect to the server. Return True if successful,
+        False if unsuccessful.
+
+        The "database" argument is used only to improve messages.
+        """
+        assert self.sqlexecute is not None
+        assert self.sqlexecute.conn is not None
+
+        # First pass with ping(reconnect=False) and minimal feedback levels.  This definitely
+        # works as expected, and is a good idea especially when "connect" was used as a
+        # synonym for "use".
+        try:
+            self.sqlexecute.conn.ping(reconnect=False)
+            if not database:
+                self.echo("Already connected.", fg="yellow")
+            return True
+        except pymysql.err.Error:
+            pass
+
+        # Second pass with ping(reconnect=True).  It is not demonstrated that this pass ever
+        # gives the benefit it is looking for, _ie_ preserves session state.  We need to test
+        # this with connection pooling.
+        try:
+            old_connection_id = self.sqlexecute.connection_id
+            self.logger.debug("Attempting to reconnect.")
+            self.echo("Reconnecting...", fg="yellow")
+            self.sqlexecute.conn.ping(reconnect=True)
+            self.logger.debug("Reconnected successfully.")
+            self.echo("Reconnected successfully.", fg="yellow")
+            self.sqlexecute.reset_connection_id()
+            if old_connection_id != self.sqlexecute.connection_id:
+                self.echo("Any session state was reset.", fg="red")
+            return True
+        except pymysql.err.Error:
+            pass
+
+        # Third pass with sqlexecute.connect() should always work, but always resets session state.
+        try:
+            self.logger.debug("Creating new connection")
+            self.echo("Creating new connection...", fg="yellow")
+            self.sqlexecute.connect()
+            self.logger.debug("New connection created successfully.")
+            self.echo("New connection created successfully.", fg="yellow")
+            self.echo("Any session state was reset.", fg="red")
+            return True
+        except pymysql.OperationalError as e:
+            self.logger.debug("Reconnect failed. e: %r", e)
+            self.echo(str(e), err=True, fg="red")
+            return False
+
     def log_output(self, output: str) -> None:
         """Log the output in the audit log, if it's enabled."""
         if isinstance(self.logfile, TextIOWrapper):
@@ -1193,10 +1279,15 @@ class MyCli:
         assert sqlexecute is not None
         assert sqlexecute.server_info is not None
         assert sqlexecute.server_info.species is not None
-        host = self.login_path if self.login_path and self.login_path_as_host else sqlexecute.host
+        if self.login_path and self.login_path_as_host:
+            prompt_host = self.login_path
+        elif sqlexecute.host is not None:
+            prompt_host = sqlexecute.host
+        else:
+            prompt_host = "localhost"
         now = datetime.now()
         string = string.replace("\\u", sqlexecute.user or "(none)")
-        string = string.replace("\\h", host or "(none)")
+        string = string.replace("\\h", prompt_host or "(none)")
         string = string.replace("\\d", sqlexecute.dbname or "(none)")
         string = string.replace("\\t", sqlexecute.server_info.species.name)
         string = string.replace("\\n", "\n")
@@ -1225,6 +1316,7 @@ class MyCli:
                 headers,
                 special.is_expanded_output(),
                 special.is_redirected(),
+                self.null_string,
             )
             for line in output:
                 click.echo(line, nl=new_line)
@@ -1239,6 +1331,7 @@ class MyCli:
                         headers,
                         special.is_expanded_output(),
                         special.is_redirected(),
+                        self.null_string,
                     )
                     for line in output:
                         click.echo(line, nl=new_line)
@@ -1250,6 +1343,7 @@ class MyCli:
         headers: list[str] | None,
         expanded: bool = False,
         is_redirected: bool = False,
+        null_string: str | None = None,
         max_width: int | None = None,
     ) -> itertools.chain[str]:
         if is_redirected:
@@ -1260,7 +1354,16 @@ class MyCli:
         expanded = expanded or use_formatter.format_name == "vertical"
         output: itertools.chain[str] = itertools.chain()
 
-        output_kwargs = {"dialect": "unix", "disable_numparse": True, "preserve_whitespace": True, "style": self.output_style}
+        output_kwargs = {
+            "dialect": "unix",
+            "disable_numparse": True,
+            "preserve_whitespace": True,
+            "style": self.output_style,
+        }
+        default_kwargs = use_formatter._output_formats[use_formatter.format_name].formatter_args
+
+        if null_string is not None and default_kwargs.get('missing_value') == DEFAULT_MISSING_VALUE:
+            output_kwargs['missing_value'] = null_string
 
         if use_formatter.format_name not in sql_format.supported_formats:
             output_kwargs["preprocessors"] = (preprocessors.align_decimals,)
@@ -1268,7 +1371,7 @@ class MyCli:
         if title:  # Only print the title if it's not None.
             output = itertools.chain(output, [title])
 
-        if cur:
+        if headers or (cur and title):
             column_types = None
             if isinstance(cur, Cursor):
 
@@ -1278,7 +1381,7 @@ class MyCli:
 
                 column_types = [get_col_type(tup) for tup in cur.description]
 
-            if max_width is not None:
+            if max_width is not None and isinstance(cur, Cursor):
                 cur = list(cur)
 
             formatted = use_formatter.format_output(
@@ -1338,7 +1441,13 @@ class MyCli:
 @click.option("--ssh-key-filename", help="Private key filename (identify file) for the ssh connection.")
 @click.option("--ssh-config-path", help="Path to ssh configuration.", default=os.path.expanduser("~") + "/.ssh/config")
 @click.option("--ssh-config-host", help="Host to connect to ssh server reading from ssh configuration.")
-@click.option("--ssl", "ssl_enable", is_flag=True, help="Enable SSL for connection (automatically enabled with other flags).")
+@click.option(
+    "--ssl-mode",
+    "ssl_mode",
+    help="Set desired SSL behavior. auto=preferred, on=required, off=off.",
+    type=click.Choice(["auto", "on", "off"]),
+)
+@click.option("--ssl/--no-ssl", "ssl_enable", default=None, help="Enable SSL for connection (automatically enabled with other flags).")
 @click.option("--ssl-ca", help="CA file in PEM format.", type=click.Path(exists=True))
 @click.option("--ssl-capath", help="CA directory.")
 @click.option("--ssl-cert", help="X509 cert in PEM format.", type=click.Path(exists=True))
@@ -1354,8 +1463,6 @@ class MyCli:
     is_flag=True,
     help=("""Verify server's "Common Name" in its cert against hostname used when connecting. This option is disabled by default."""),
 )
-# as of 2016-02-15 revocation list is not supported by underling PyMySQL
-# library (--ssl-crl and --ssl-crlpath options in vanilla mysql client)
 @click.version_option(__version__, "-V", "--version", help="Output mycli's version.")
 @click.option("-v", "--verbose", is_flag=True, help="Verbose output.")
 @click.option("-D", "--database", "dbname", help="Database to use.")
@@ -1372,7 +1479,9 @@ class MyCli:
     is_flag=True,
     help="Automatically switch to vertical output mode if the result is wider than the terminal width.",
 )
-@click.option("--show-warnings/--no-show-warnings", is_flag=True, help="Automatically show warnings after executing a SQL statement.")
+@click.option(
+    "--show-warnings/--no-show-warnings", "show_warnings", is_flag=True, help="Automatically show warnings after executing a SQL statement."
+)
 @click.option("-t", "--table", is_flag=True, help="Display batch output in table format.")
 @click.option("--csv", is_flag=True, help="Display batch output in CSV format.")
 @click.option("--warn/--no-warn", default=None, help="Warn before running a destructive query.")
@@ -1402,6 +1511,7 @@ def cli(
     auto_vertical_output: bool,
     show_warnings: bool,
     local_infile: bool,
+    ssl_mode: str | None,
     ssl_enable: bool,
     ssl_ca: str | None,
     ssl_capath: str | None,
@@ -1448,6 +1558,15 @@ def cli(
         warn=warn,
         myclirc=myclirc,
     )
+
+    if ssl_enable is not None:
+        click.secho(
+            "Warning: The --ssl/--no-ssl CLI options are deprecated and will be removed in a future release. "
+            "Please use the ssl_mode config or --ssl-mode CLI options instead.",
+            err=True,
+            fg="yellow",
+        )
+
     if list_dsn:
         try:
             alias_dsn = mycli.config["alias_dsn"]
@@ -1544,19 +1663,36 @@ def cli(
             ssl_verify_server_cert = ssl_verify_server_cert or (params[0].lower() == 'true')
             ssl_enable = True
 
-    ssl = {
-        "enable": ssl_enable,
-        "ca": ssl_ca and os.path.expanduser(ssl_ca),
-        "cert": ssl_cert and os.path.expanduser(ssl_cert),
-        "key": ssl_key and os.path.expanduser(ssl_key),
-        "capath": ssl_capath,
-        "cipher": ssl_cipher,
-        "tls_version": tls_version,
-        "check_hostname": ssl_verify_server_cert,
-    }
+    ssl_mode = ssl_mode or mycli.ssl_mode  # cli option or config option
 
-    # remove empty ssl options
-    ssl = {k: v for k, v in ssl.items() if v is not None}
+    # if there is a mismatch between the ssl_mode value and other sources of ssl config, show a warning
+    # specifically using "is False" to not pickup the case where ssl_enable is None (not set by the user)
+    if ssl_enable and ssl_mode == "off" or ssl_enable is False and ssl_mode in ("auto", "on"):
+        click.secho(
+            f"Warning: The current ssl_mode value of '{ssl_mode}' is overriding the value provided by "
+            f"either the --ssl/--no-ssl CLI options or a DSN URI parameter (ssl={ssl_enable}).",
+            err=True,
+            fg="yellow",
+        )
+
+    # configure SSL if ssl_mode is auto/on or if
+    # ssl_enable = True (from --ssl or a DSN URI) and ssl_mode is None
+    if ssl_mode in ("auto", "on") or (ssl_enable and ssl_mode is None):
+        ssl = {
+            "mode": ssl_mode,
+            "enable": ssl_enable,
+            "ca": ssl_ca and os.path.expanduser(ssl_ca),
+            "cert": ssl_cert and os.path.expanduser(ssl_cert),
+            "key": ssl_key and os.path.expanduser(ssl_key),
+            "capath": ssl_capath,
+            "cipher": ssl_cipher,
+            "tls_version": tls_version,
+            "check_hostname": ssl_verify_server_cert,
+        }
+        # remove empty ssl options
+        ssl = {k: v for k, v in ssl.items() if v is not None}
+    else:
+        ssl = None
 
     if ssh_config_host:
         ssh_config = read_ssh_config(ssh_config_path).lookup(ssh_config_host)
